import { call, put, takeEvery } from 'redux-saga/effects'

/** The basic CRUD API supports the following functionality:
  * /api[/module][/id|/offset[/limit]]
  *
  * Other functions
  */

import request from 'request-promise'
export function crudHttpApi (action, options) {
  const { uri } = options
  // determine the operation
  const [ name, actionType ] = action.type.split('/')
  const requestOptions = {
    uri: `${uri}/${name}`,
    method: null,
    body: null
  }
  switch (actionType) {
    case 'READ_REQUEST':
      options.uri = `${uri}/${action.id}`
      options.method = 'GET'
      break
    case 'CREATE_REQUEST':
      options.method = 'POST'
      options.body = action.data
      break
    case 'UPDATE_REQUEST':
      options.uri = `${uri}/${action.id}`
      options.method = 'PUT'
      break
    case 'DELETE_REQUEST':
      options.uri = `${uri}/${action.id}`
      options.method = 'DELETE'
      break
  }
  // Request from the request-promise library already returns a promise.
  return request(requestOptions)
}

export function crudLocalstorageApi (action, options) {
  // determine the operation
  const [ name, actionType ] = action.type.split('/')
  switch (actionType) {
    case 'READ_REQUEST':
      return new Promise((resolve, reject) => {
        const items = localStorage.getItem(name) || []
        resolve(items[action.id])
      })
    case 'CREATE_REQUEST':
      return new Promise((resolve, reject) => {
        try {
          const items = localStorage.getItem(name) || {}
          items[action.id] = action.data
          console.log(items)
          localStorage.setItem(name, items)
          resolve(null)
        } catch (error) {
          reject(error)
        }
      })
    case 'UPDATE_REQUEST':
      return new Promise((resolve, reject) => {
        try {
          const items = localStorage.getItem(name) || {}
          items[action.id] = action.data
          localStorage.setItem(name, items)
          resolve(null)
        } catch (error) {
          reject(error)
        }
      })
    case 'DELETE_REQUEST':
      return new Promise((resolve, reject) => {
        try {
          const items = localStorage.getItem(name) || {}
          delete items[action.id]
          localStorage.setItem(name, items)
          resolve(null)
        } catch (error) {
          reject(error)
        }
      })
  }
}

/** reduxCrudGen
  *
  * Generates CRUD action types, actions, reducers, workers and watchers.
  */
export default function reduxCrudGen (name, prefix, api) {
  const actionTypes = {}
  const actions = {}

  /** Supported actions */
  const actionList = ['create', 'read', 'update', 'delete']

  /** For each supported action create an action creator to request
    * data from the API, a success action when data is processed, and
    * a failure action, when the API call was not successful.
    * You will get e.g. create_request, update_success... */
  actionList.forEach(action => {
    /** generate sth like 'createProject' which dispatches 'project/CREATE_REQUEST' */
    const actionName = `${action}${name[0].toUpperCase()}${name.substring(1)}`
    if (action === 'list') { actions[actionName] = (data, id = null) => { return { type: `${name}/${action}_REQUEST`, id, data } } }
    ['request', 'success', 'failure'].forEach(mode => {
      /** generate sth like 'project/CREATE_REQUEST' */
      const actionType = `${action}_${mode}`.toUpperCase()
      actionTypes[actionType] = `${name}/${actionType}`
    })
  })
  actions[`create${name[0].toUpperCase() + name.substring(1)}`] = (data) => { return { type: `${name}/CREATE_REQUEST`, data } }
  actions[`read${name[0].toUpperCase() + name.substring(1)}`] = (id, data) => { return { type: `${name}/READ_REQUEST`, id, data } }
  actions[`update${name[0].toUpperCase() + name.substring(1)}`] = (id, data) => { return { type: `${name}/UPDATE_REQUEST`, id, data } }
  actions[`delete${name[0].toUpperCase() + name.substring(1)}`] = (id) => { return { type: `${name}/DELETE_REQUEST`, id } }

  const state = []

  function * worker (action) {
    try {
      const data = yield call(api, action)
      yield put({ type: action.type.replace('REQUEST', 'SUCCESS'), data })
    } catch (error) {
      yield put({ type: action.type.replace('REQUEST', 'FAILURE'), error })
    }
  }

  /** The watcher function hijacks the _REQUEST actions and calls
    * asynchronous workers. Once they yield their result, a _SUCCESS
    * or _FAILURE action is triggered. */
  function * watcher () {
    let action
    for (action of actionList) {
      yield takeEvery(`${name}/${action.toUpperCase()}_REQUEST`, worker)
    }
  }

  function reducer (state = [], action) {
    let nextState
    if (!action) {
      return state
    }
    switch (action.type) {
      case actionTypes[`CREATE_SUCCESS`]:
        nextState = [ ...state, action.data ]
        return nextState
      case actionTypes[`UPDATE_SUCCESS`]:
        if (typeof action.id !== 'undefined' && action.id >= 0) {
          nextState = [ ...state.slice(0, action.id), { ...state[action.id], ...action.data }, ...state.slice(action.id + 1) ]
          return nextState
        } else {
          return state
        }
      case actionTypes[`DELETE_SUCCESS`]:
        if (typeof action.id !== 'undefined' && action.id >= 0) {
          nextState = [ ...state.slice(0, action.id), ...state.slice(action.id + 1) ]
          return nextState
        } else {
          return state
        }
      default:
        return state
    }
  }

  return {
    actionTypes,
    actions,
    state,
    reducer,
    worker,
    watcher
  }
}
